/* ***** BEGIN LICENSE BLOCK *****
* Version: MPL 1.1/GPL 2.0/LGPL 2.1
*
* The contents of this file are subject to the Mozilla Public License Version
* 1.1 (the "License"); you may not use this file except in compliance with
* the License. You may obtain a copy of the License at
* http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
* for the specific language governing rights and limitations under the
* License.
*
* The Original Code is Bookmarks Menu Button.
*
* The Initial Developer of the Original Code is Ronny Perinke.
* Portions created by the Initial Developer are Copyright (C) 2004-2009
* the Initial Developer. All Rights Reserved.
*
* Contributor(s):
*
*   Ronny Perinke <www.sephiroth-j.de>
*
* Alternatively, the contents of this file may be used under the terms of
* either the GNU General Public License Version 2 or later (the "GPL"), or
* the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
* in which case the provisions of the GPL or the LGPL are applicable instead
* of those above. If you wish to allow use of your version of this file only
* under the terms of either the GPL or the LGPL, and not to allow others to
* use your version of this file under the terms of the MPL, indicate your
* decision by deleting the provisions above and replace them with the notice
* and other provisions required by the GPL or the LGPL. If you do not delete
* the provisions above, a recipient may use your version of this file under
* the terms of any one of the MPL, the GPL or the LGPL.
*
* ***** END LICENSE BLOCK ***** */

const bmbextHistory = {
 	//#######################################
	//########	private functions and properties	########
	//########						########
	//#######################################
	_Prefs:
	{
		hideDropmarker: false,				// hide the dropmarker?
		hideHistoryMenu: false,				// hide normal history menu?
		hideBMBS: false,					// remove BMB-Settings from history button?
		MiddleClickOpens: 1,				// middle-click on button loads ...
		RightClickOpens: 1,				// right-click on button loads ...
		GroupMethod: 2,					// how to group the history
		MaxRecentHistory: 10,				// max recently visited sites
		enableDebug: false				// enable debug output to console and error console
	},
	_MiscVars:
	{
		ShortcutKey: null,
		ShortcutKeyPressed: false,			// was the shortcut key pressed?
		Selection: null,
		PrefsIFace: null,					// refers to the nsIPref interface
		bmbCSSSheet:					// container for the CSS-Style sheet of BMB
		{
			globalIndex: -1,				// its index in documents.styleSheets
			sheetObj: null,				// the styleSheet object itself
			dropmarkerHidden: false		// did we hide the dropmarker?
		},
		TabImages: [],					// array that will contain the image-uri of a closed tab or empty string if no image
		canUndoCloseWin: (typeof(undoCloseWindow) === "function")
	},	
	get prefRoot()
	{
		return "BMBExt.History.";
	},
	get prefRootG()
	{
		return "BMBExt.";
	},
	get StringBundle()
	{
		var stringBundle = document.getElementById("bmbextstr");
		delete this.StringBundle;
		this.StringBundle = stringBundle;
		return stringBundle;
	},
	get PrefsIFace()
	{
		if (!this._MiscVars.PrefsIFace)
		{
			this.PrefsIFace = this.prefRoot;
			this.__defineGetter__("PrefsIFace", function() { return this._Misc.PrefsIFace; });
		}
		return this._MiscVars.PrefsIFace;
	},
	set PrefsIFace(aBranch)
	{
		if (typeof(aBranch) == "string")
		{
			this._MiscVars.PrefsIFace = Components.classes["@mozilla.org/preferences;1"].getService(Components.interfaces.nsIPref);
			this._MiscVars.PrefsIFace = this._MiscVars.PrefsIFace.getBranch(aBranch?aBranch:this.prefRoot);
		}
		else
		{
			throw Components.results.NS_ERROR_ILLEGAL_VALUE + "\nValue must be a string type";
		}
		
	},
	get Selection()
	{
		return this._MiscVars.Selection;
	},
	set Selection(aObject)
	{
		this._MiscVars.Selection = aObject;
	},
	get BMBCSSSheet()
	{
		return this._MiscVars.bmbCSSSheet;
	},
	HistoryGroupingStringToInt: function BMBH_HistoryGroupingStringToInt(aGroupMethod)
	{
		if (typeof(aGroupMethod) != "string")
			return null;
		
		switch(aGroupMethod)
		{
			case "place:sort=0&maxResults=100":
				return 0;
			
			case "place:type=" + Components.interfaces.nsINavHistoryQueryOptions.RESULTS_AS_DATE_QUERY:
				return 1;
			
			case "place:type=" + Components.interfaces.nsINavHistoryQueryOptions.RESULTS_AS_DATE_SITE_QUERY:
				return 2;
			
			case "place:type=" + Components.interfaces.nsINavHistoryQueryOptions.RESULTS_AS_DATE_DOMAIN_SITE_QUERY:
				return 3;
			
			default:
				return -1;
		}
	},
	HistoryGroupingIntToString: function BMBH_HistoryGroupingIntToString(aGroupMethod)
	{
		if (typeof(aGroupMethod) != "number")
			return null;
		
		switch(aGroupMethod)
		{
			case 0:
				return "place:sort=0&maxResults=100";
				
			case 1:
				return "place:type=" + Components.interfaces.nsINavHistoryQueryOptions.RESULTS_AS_DATE_QUERY;
				
			case 2:
			default:
				return "place:type=" + Components.interfaces.nsINavHistoryQueryOptions.RESULTS_AS_DATE_SITE_QUERY;
				
			case 3:
				return "place:type=" + Components.interfaces.nsINavHistoryQueryOptions.RESULTS_AS_DATE_DOMAIN_SITE_QUERY;
		}
	},
	get ConsService()
	{
		var consService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
		delete this.ConsService;
		this.ConsService = consService;
		return consService;
	},
	dump: function BHMB_dump(aMsg, aLine, aType)
	{
		if (!this._Prefs.enableDebug)
		{
			return;
		}
		else
		{
			dump("BookmarkMenuButton::" + aMsg + "\n");
			if (!isNaN(aType) && aType != 3)
			{
				var scriptError = Components.classes["@mozilla.org/scripterror;1"].createInstance(Components.interfaces.nsIScriptError);
				//nsIScriptError.errorFlag = 0, nsIScriptError.warningFlag = 1, nsIScriptError.exceptionFlag = 2 and nsIScriptError.strictFlag = 4.
				scriptError.init("BookmarkMenuButton::" + aMsg, "chrome://bmb-ext/content/bmb-ext_History.js", null, aLine, 0, aType, "XUL javascript");
				this.ConsService.logMessage(scriptError);
			}
			else
			{
				this.ConsService.logStringMessage("BookmarkMenuButton::" + aMsg);
			}
		}
	},
	//#######################################
	//########	/private functions and properties	########
	//########						########
	//#######################################
	
	/**
		implement EventListener Interface
		only used for addEventListener()
	*/
	handleEvent: function BMBH_handleEvent(aEvent)
	{
		switch(aEvent.type)
		{
			case "load":
				this.init();
				break;
			
			case "unload":
				this.quit();
				break;
			
			case "keydown":
			case "keyup":
				this.catchKKey(aEvent);
				break;
			
			case "TabClose":
				this.tabClosed(aEvent);
				break;
		}
	},
	
	init: function BMBH_init()
	{
		this.loadGPrefs();
		this.loadPrefs();
		
		this.hideHistoryMenu();
		this.findBMBCSSSheet();
		this.showHideDropmarker();
		var button = document.getElementById("history-menu-button");
		if (button)
		{
			this.checkForContextMenu();
			this.addBMBSettings();
			this.setHistoryGrouping();
			
			// SHORTCUT-KEY
			this._MiscVars.ShortcutKey = document.getElementById("history-menu").getAttribute("accesskey").charCodeAt(0);		//converts the character to the ASCII code because keyCode is a ASCII code
			window.addEventListener("keydown", this, false);
			window.addEventListener("keyup", this, false);
		}
		
		// safe the image of the tab when it gets closed
		document.getElementById("content").addEventListener("TabClose", this, false);
	},
	quit: function BMBH_quit()
	{
		this._MiscVars.bmbCSSSheet.sheetObj = null;
		this._MiscVars.PrefsIFace = null;
		this._MiscVars.TabImages = null;
		delete this.ConsService;
		delete this.StringBundle;
		this.Selection = null;
	},
	
	loadPrefs: function BMBH_loadPrefs()
	{
		this.PrefsIFace = this.prefRoot;
		
		this._Prefs.hideDropmarker = this.PrefsIFace.getBoolPref("dropmarker.hide");
		this._Prefs.hideHistoryMenu = this.PrefsIFace.getBoolPref("historymenu.hide");
		this._Prefs.hideBMBS = this.PrefsIFace.getBoolPref("settings.hide");
		this._Prefs.MiddleClickOpens = this.PrefsIFace.getIntPref("settings.MiddleClickOpens");
		this._Prefs.RightClickOpens = this.PrefsIFace.getIntPref("settings.RightClickOpens");
		this._Prefs.GroupMethod = this.PrefsIFace.getIntPref("settings.GroupMethod");
		this._Prefs.MaxRecentHistory = this.PrefsIFace.getIntPref("settings.MaxRecentHistory");
	},
	loadGPrefs: function BMB_loadPrefs()
	{
		this.PrefsIFace = this.prefRootG;
		
		this._Prefs.enableDebug = this.PrefsIFace.getBoolPref("enableDebug");
	},
	
	hideHistoryMenu: function BMBH_hideHistoryMenu()
	{
		// only hide it, when hmb is used an the pref is true
		document.getElementById("history-menu").hidden = document.getElementById("history-menu-button") && this._Prefs.hideHistoryMenu;
	},
	findBMBCSSSheet: function BMBH_findBMBCSSSheet()
	{
		var CSSSheets = document.styleSheets;
		for (let i=0, l=CSSSheets.length; i<l; i++)
		{
			if (CSSSheets[i].href == "chrome://bmb-ext/skin/bmb-ext.css")
			{
				this.BMBCSSSheet.globalIndex = i;
				this.BMBCSSSheet.sheetObj = CSSSheets[i];
				break;
			}
		}
	},
	showHideDropmarker: function BMBH_showHideDropmarker()
	{
		var sheet = this.BMBCSSSheet;
		if (sheet.globalIndex < 0)	// did not found it
			return;
		
		var lastRuleIndex = sheet.sheetObj.cssRules.length - 1;
		
		// hide it
		if (!sheet.dropmarkerHidden && this._Prefs.hideDropmarker)
		{
			sheet.sheetObj.insertRule("#history-menu-button .toolbarbutton-menu-dropmarker { display: none !important;}", lastRuleIndex + 1);
			sheet.dropmarkerHidden = true;
			return;	// sind fertig mit verstecken, wollen es ja nicht gleich wieder anzeigen ;-)
		}
		
		// unhide it
		if (sheet.dropmarkerHidden && !this._Prefs.hideDropmarker)
		{
			var indexToRemove = -1;
			for (let i = lastRuleIndex; i >= 0; i--)
			{
				if (sheet.sheetObj.cssRules[i].selectorText == "#history-menu-button .toolbarbutton-menu-dropmarker")
				{
					indexToRemove = i;
					break;
				}
			}
			
			if (indexToRemove >= 0)
			{
				sheet.sheetObj.deleteRule(indexToRemove);
				sheet.dropmarkerHidden = false;
			}
		}
	},
	
	/**
		remove/disable toolbar context menu, if needed (right-click perfoms user-action)
	*/
	checkForContextMenu: function BMBH_checkForContextMenu()
	{
		var button = document.getElementById("history-menu-button");
		
		if (this._Prefs.RightClickOpens > 0)
		{
			button.setAttribute("context", "");
		}
		else
		{
			button.setAttribute("context", "toolbar-context-menu");
		}
	},
	
	onMouseOverButton: function BMBH_onMouseOverButton(aEvent)
	{
		if (aEvent.target.id == "history-menu-button")
		{
			this.checkForContextMenu();
			this.addBMBSettings();
			this.setHistoryGrouping();
		}
	},
	onPopupShowing: function BMBH_onPopupShowing(aEvent, popupNumber)
	{
		if (popupNumber == 1 && aEvent.target.id == "history-menu-button-popup")
		{
			this.checkForWinUndo();
			this.updateTabUndoMenu();
			this.updateWinUndoMenu();
			this.updateRecentHistory_Places(aEvent.target);
			this.updateSanitizeCommand();
		}
		if (popupNumber == 2)
		{
			this.addOpenAllInTabs(aEvent.target);
		}
	},
	onButtonCommand: function BMBH_onButtonCommand(aEvent, closeMenu)
	{
		var target = aEvent.target;
		if (target.parentNode && (target.parentNode.id == "history-menu-button-TabUndoMenuPopup" || target.parentNode.id == "history-menu-button-WinUndoMenuPopup"))
		{
			return;
		}
		
		var url = target.node?target.node.uri:"";
		if (url)
		{
			openUILink(url, aEvent, false, true);
			if (closeMenu)
			{
				closeMenus(target);
			}
			return;
		}
		
		if (target.hasAttribute("openInTabs"))
		{
			PlacesUIUtils.openContainerNodeInTabs(target.parentNode._resultNode, aEvent);
			return;
		}
	},
	onButtonClick: function BMBH_onButtonClick(aEvent, aNode)
	{
		this.dump("bmbextHistory.onButtonClick");
		this.catchMKey(aEvent, aNode);
		this.checkForMiddleClick(aNode, aEvent);
	},
	checkForMiddleClick: function BMBH_checkForMiddleClick(aNode, aEvent)
	{
		if (aEvent.button == 1)
		{
			this.onButtonCommand(aEvent, true);
		}
	},
	
	/**
		adds/removes BMB settings menuitem
	*/
	addBMBSettings: function BMBH_addBMBSettings()
	{
		var settings = document.getElementById("hmb-settings");
		var settingssep = document.getElementById("hmb-settings-sep");
		settings.hidden = this._Prefs.hideBMBS;
		settingssep.hidden = this._Prefs.hideBMBS;
	},
	
	/**
		changes the grouping of the total history, if needed
	*/
	setHistoryGrouping: function BMBH_setHistoryGrouping()
	{
		var HMBHpopup = document.getElementById("history-menu-button-history-popup").firstChild;
		if (this.HistoryGroupingStringToInt(HMBHpopup.place) != this._Prefs.GroupMethod)
		{
			HMBHpopup.place = this.HistoryGroupingIntToString(this._Prefs.GroupMethod);
		}
	},
	
	/**
		performs user-action on right/middle click on the (closed) button
	*/
	catchMKey: function BMBH_catchMKey(aEvent, aNode)
	{
		//aEvent.currentTarget does not work correctly. if it's used, than middle click will toggle sidebar even if the button is/was open, because open-state is already false
		if (aEvent.target.id != "history-menu-button" || aEvent.currentTarget.disabled)
			return;
		
		//this.dump("bmbextHistory.catchMKey");
		var hmb = aEvent.target;
		switch(aEvent.button)
		{
			case 0:	//left button; uses oncommand
				break;
				
			case 1:	//middle button
				if (!hmb.open)
				{
					switch(this._Prefs.MiddleClickOpens)
					{
						case 0:
						default:
							break;
						case 1:
							toggleSidebar("viewHistorySidebar");
							break;
						case 2:
							(this._MiscVars.canUndoCloseWin && aEvent.ctrlKey) ? this.openLastClosedWindow() : this.openLastClosedTab();
							break;
						case 3:
							for (let i = 0, l = gBrowser.browsers.length; i < l; i++)
							{
								if (gBrowser.browsers[i].contentDocument.location == "about:history")
								{
									//gBrowser.selectTabAtIndex(i);	// nur ab Fx 3.5
									gBrowser.selectedTab = gBrowser.mTabs[i];
									return;
								}
							}
							let where = (gBrowser.browsers.length == 1 && gBrowser.browsers[0].contentDocument.location == "about:blank") ? "current" : "tab";
							openUILinkIn("about:history", where);
							break;
					}
				}
				break;
				
			case 2:	//right button
				if (!hmb.open)
				{
					switch(this._Prefs.RightClickOpens)
					{
						case 0:
						default:
							break;
						case 1:
							toggleSidebar("viewHistorySidebar");
							break;
						case 2:
							(this._MiscVars.canUndoCloseWin && aEvent.ctrlKey) ? this.openLastClosedWindow() : this.openLastClosedTab();
							break;
						case 3:
							for (let i = 0, l = gBrowser.browsers.length; i < l; i++)
							{
								if (gBrowser.browsers[i].contentDocument.location == "about:history")
								{
									//gBrowser.selectTabAtIndex(i);	// nur ab Fx 3.5
									gBrowser.selectedTab = gBrowser.mTabs[i];
									return;
								}
							}
							let where = (gBrowser.browsers.length == 1 && gBrowser.browsers[0].contentDocument.location == "about:blank") ? "current" : "tab";
							openUILinkIn("about:history", where);
							break;
					}
				}
				break;
				
			default:	//unknown
				break;
		}
	},
	/**
		handles shortcut-key events and opens the button after releasing the shortcut-key
	*/
	catchKKey: function BMBH_catchKKey(aEvent)
	{
		//shortcut key event handler
		var hmb = document.getElementById("history-menu-button");
		if (!hmb || hmb.open)
		{
			this._MiscVars.ShortcutKeyPressed = false;	//zur sicherheit
			return;
		}
		
		if (!this._MiscVars.ShortcutKey)
		{
			return;
		}
		
		if (aEvent.type== "keydown" && !this._MiscVars.ShortcutKeyPressed)
		{
			if (aEvent.altKey && (aEvent.keyCode == this._MiscVars.ShortcutKey))
			{
				this._MiscVars.ShortcutKeyPressed = true;
			}
		}
		if (aEvent.type== "keyup" && this._MiscVars.ShortcutKeyPressed)
		{
			if (aEvent.altKey && (aEvent.keyCode == this._MiscVars.ShortcutKey))
			{
				this._MiscVars.ShortcutKeyPressed = false;
				this.openMenu();
			}
		}
	},
	openMenu: function BMBH_openMenu()
	{
		this.checkForContextMenu();
		this.addBMBSettings();
		this.setHistoryGrouping();
		
		var hmb = document.getElementById("history-menu-button");
		hmb.firstChild.openPopup(hmb, "after_start", 0, 0, false, false);
	},
	
	onHistoryButtonHidden: function()
	{
	},
	updateRecentHistory_Places: function BMBH_updateRecentHistory_Places(HistoryButton) 
	{
		var place = HistoryButton.place.replace(/maxResults=[0-9]+/, "maxResults=" + this._Prefs.MaxRecentHistory);
		HistoryButton.place = place;
		
		var endSep = document.getElementById("endHistorySeparator2");
		var resultNode = HistoryButton.getResultNode();
		var wasOpen = resultNode.containerOpen;
		resultNode.containerOpen = true;
		endSep.hidden = resultNode.childCount == 0;
		if (!wasOpen)
		{
			resultNode.containerOpen = false;
		}
	},
	/**
		adds "Open All In Tabs" command at the bottom of each menu in the total history menu, if it has at least 2 menuitems
	*/
	addOpenAllInTabs: function BMBH_addOpenAllInTabs(menupopup)
	{
		if (	menupopup && menupopup.childNodes.length > 1 && 
			!menupopup.lastChild.hasAttribute("openInTabs") && menupopup.getElementsByTagName("menuitem").length > 1
		)
		{
			var openInTabs = document.createElement("menuitem");
			openInTabs.setAttribute("openInTabs", "true");
			openInTabs.setAttribute("label", gNavigatorBundle.getString("menuOpenAllInTabs.label"));
			
			menupopup.appendChild(document.createElement("menuseparator"));
			menupopup.appendChild(openInTabs);
		}
	},
	
	/*
		SANITIZE CODE	-	the clean and simple way
	*/
	updateSanitizeCommand: function BMBH_updateSanitizeCommand()
	{
		var completeHistory = document.getElementById("history-menu-button-history-popup");
		var SanitizeHistoryItem = document.getElementById("SanitizeHistory");
		var active = false;	//state of item - false by default
		var historyService = Components.classes["@mozilla.org/browser/nav-history-service;1"].getService(Components.interfaces.nsINavHistoryService);
		active = historyService.hasHistoryEntries;	// http://lxr.mozilla.org/mozilla/source/toolkit/components/places/public/nsINavHistoryService.idl#1112
		
		SanitizeHistoryItem.setAttribute("disabled", !active);
		completeHistory.setAttribute("disabled", !active);
	},
	doSanitizeHistory: function BMBH_doSanitizeHistory()
	{
		setTimeout("bmbextHistory.doRealSanitizeHistory()", 1);
	},
	doRealSanitizeHistory: function BMBH_doRealSanitizeHistory()
	{
		var globalHistory = Components.classes["@mozilla.org/browser/global-history;2"].getService(Components.interfaces.nsIBrowserHistory);
		globalHistory.removeAllPages();
		
		var observer = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
		try
		{
			observer.notifyObservers(null, "browser:purge-session-history", "");
		}
		catch(e)
		{}
	},
	
	/*
		TAB/WINDOW UNDO MENU and extra functions
	*/
	checkForWinUndo: function BMBH_checkForWinUndo()
	{
		if (!this._MiscVars.canUndoCloseWin)
			return;
		
		var menuUndo = document.getElementById("historyUndoWindowMenu");
		if (menuUndo)
		{
			var buttonUndoMenu = document.getElementById("history-menu-button-WinUndoMenu");
			buttonUndoMenu.setAttribute("label", menuUndo.getAttribute("label"));
			buttonUndoMenu.hidden = false;
			// wird in updateWinUndoMenu() enabled, wenn es fenster gibt
		}
	},
	updateTabUndoMenu: function BMBH_updateTabUndoMenu()
	{
		var buttonUndoMenu = document.getElementById("history-menu-button-TabUndoMenu");
		
		var undoPopup = buttonUndoMenu.firstChild;
		// remove existing menu items
		while (undoPopup.hasChildNodes())
		{
			undoPopup.removeChild(undoPopup.firstChild);
		}
		
		// get closed-tabs from nsSessionStore
		var ss = Components.classes["@mozilla.org/browser/sessionstore;1"].getService(Components.interfaces.nsISessionStore);
		// no restorable tabs, so make sure menu is disabled, and return
		if (ss.getClosedTabCount(window) == 0)
		{
			buttonUndoMenu.setAttribute("disabled", true);
			return;
		}
		
		// enable menu
		buttonUndoMenu.removeAttribute("disabled");
		
		// populate menu
		var undoItems = eval("(" + ss.getClosedTabData(window) + ")");
		var menuitem;
		for (let i=0, l=undoItems.length, item, itemTitle, itemURL; i < l; i++)
		{
			item = undoItems[i];
			itemTitle = item.title;
			itemURL = item.state.entries[item.state.index - 1].url;
			menuitem = document.createElement("menuitem");
			menuitem.setAttribute("label", itemTitle);
			menuitem.setAttribute("oncommand", "bmbextHistory.openClosedTabX(" + i + ");");
			menuitem.setAttribute("tooltiptext", itemTitle);
			menuitem.setAttribute("statustext", itemURL);
			menuitem.setAttribute("onmouseover", "bmbext.setStatus(this);");
			menuitem.setAttribute("class", "menuitem-iconic tabitem");
			let imgsrc = item.image||this._MiscVars.TabImages[i];	// new since bug 389487
			if (imgsrc && !gBrowser.isFailedIcon(imgsrc))	// isFailedIcon() -> http://lxr.mozilla.org/mozilla/source/browser/base/content/tabbrowser.xml#572
			{
				// don't initiate a connection just to fetch a favicon (see bug 467828)
				if (/^https?:/.test(imgsrc))
				{
					imgsrc = "moz-anno:favicon:" + imgsrc;
				}
				menuitem.setAttribute("image", imgsrc);
			}
			undoPopup.appendChild(menuitem);
		}
		// safe one more if
		// I can do this safely here, because if there are no tabs to restore the function wouldn't even get to this point of the code
		undoPopup.firstChild.setAttribute("key", "key_undoCloseTab");	// new since bug 357235
		let undoItemsLength = undoItems.length;
		undoItems = null;
		
		// "open all in tabs"
		var bundleService = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
		var bmStringBundle = bundleService.createBundle("chrome://browser/locale/browser.properties");
		undoPopup.appendChild(document.createElement("menuseparator"));
		menuitem = undoPopup.appendChild(document.createElement("menuitem"));
		menuitem.setAttribute("label", bmStringBundle.GetStringFromName("menuOpenAllInTabs.label"));
		menuitem.setAttribute("accesskey", bmStringBundle.GetStringFromName("menuOpenAllInTabs.accesskey"));
		menuitem.setAttribute("oncommand", "bmbextHistory.openClosedTabs(" + undoItemsLength + ");");
		menuitem.setAttribute("class", "menuitem-iconic openall");
		
		// "clear list"
		menuitem = undoPopup.appendChild(document.createElement("menuitem"));
		menuitem.setAttribute("label", this.StringBundle.getString("bmb-ext.cleartablist.label"));
		menuitem.setAttribute("accesskey", this.StringBundle.getString("bmb-ext.cleartablist.accesskey"));
		menuitem.setAttribute("oncommand", "bmbextHistory.clearClosedTabs();");
		menuitem.setAttribute("class", "menuitem-iconic cleartabs");
	},
	updateWinUndoMenu: function BMBH_updateTabUndoMenu()
	{
		var buttonWinUndoMenu = document.getElementById("history-menu-button-WinUndoMenu");
		if (!this._MiscVars.canUndoCloseWin || buttonWinUndoMenu.hidden)
		{
			return;
		}
		
		var undoPopup = buttonWinUndoMenu.firstChild;
		// remove existing menu items
		while (undoPopup.hasChildNodes())
		{
			undoPopup.removeChild(undoPopup.firstChild);
		}
		
		// get closed-tabs from nsSessionStore
		var ss = Components.classes["@mozilla.org/browser/sessionstore;1"].getService(Components.interfaces.nsISessionStore);
		// no restorable tabs, so make sure menu is disabled, and return
		if (ss.getClosedWindowCount(window) == 0)
		{
			buttonWinUndoMenu.setAttribute("disabled", true);
			return;
		}
		
		// enable menu
		buttonWinUndoMenu.removeAttribute("disabled");
		
		var bundleService = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
		var bmStringBundle = bundleService.createBundle("chrome://browser/locale/browser.properties");
		var menuLabelString = bmStringBundle.GetStringFromName("menuUndoCloseWindowLabel");
		var menuLabelStringSingleTab = bmStringBundle.GetStringFromName("menuUndoCloseWindowSingleTabLabel");
		
		// populate menu
		var undoItems = JSON.parse(ss.getClosedWindowData());
		for (let i=0, l=undoItems.length, item; i < l; i++)
		{
			item = undoItems[i];
			let menuitem = document.createElement("menuitem");
			let itemNumOtherTabs = item.tabs.length -1 ;
			let itemSelectedTab = item.tabs[item.selected - 1];
			let label;
			if (itemNumOtherTabs == 0)
			{
				label = menuLabelStringSingleTab;
				// url of first (and only) tab
				menuitem.setAttribute("statustext", itemSelectedTab.entries[0].url);
				menuitem.setAttribute("onmouseover", "bmbext.setStatus(this);");
			}
			else
			{
				label = PluralForm.get(itemNumOtherTabs, menuLabelString);
			}
			let itemTitle = label.replace("#1", item.title).replace("#2", itemNumOtherTabs);
			menuitem.setAttribute("label", itemTitle);
			menuitem.setAttribute("oncommand", "bmbextHistory.openClosedWindowX(" + i + ");");
			menuitem.setAttribute("tooltiptext", itemTitle);
			menuitem.setAttribute("class", "menuitem-iconic winitem");
			let imgsrc = itemSelectedTab.attributes.image;
			if (imgsrc && !gBrowser.isFailedIcon(imgsrc))	// isFailedIcon() -> http://mxr.mozilla.org/mozilla1.9.1/source/browser/base/content/tabbrowser.xml#724
			{
				// don't initiate a connection just to fetch a favicon (see bug 467828)
				if (/^https?:/.test(imgsrc))
				{
					imgsrc = "moz-anno:favicon:" + imgsrc;
				}
				menuitem.setAttribute("image", imgsrc);
			}
			undoPopup.appendChild(menuitem);
		}
		// I can do this safely here, because if there are no tabs to restore the function wouldn't even get to this point of the code
		undoPopup.firstChild.setAttribute("key", "key_undoCloseWindow");
		let undoItemsLength = undoItems.length;
		undoItems = null;
		
		// "open all in tabs"
		undoPopup.appendChild(document.createElement("menuseparator"));
		var menuitem = undoPopup.appendChild(document.createElement("menuitem"));
		menuitem.setAttribute("label", bmStringBundle.GetStringFromName("menuRestoreAllWindows.label"));
		menuitem.setAttribute("accesskey", bmStringBundle.GetStringFromName("menuRestoreAllWindows.accesskey"));
		menuitem.setAttribute("oncommand", "bmbextHistory.openClosedWindows(" + undoItemsLength + ");");
		menuitem.setAttribute("class", "menuitem-iconic openall");
		
		// "clear list"
		menuitem = undoPopup.appendChild(document.createElement("menuitem"));
		menuitem.setAttribute("label", this.StringBundle.getString("bmb-ext.clearwindowlist.label"));
		menuitem.setAttribute("accesskey", this.StringBundle.getString("bmb-ext.clearwindowlist.accesskey"));
		menuitem.setAttribute("oncommand", "bmbextHistory.clearClosedWindows();");
		menuitem.setAttribute("class", "menuitem-iconic cleartabs");
	},
	openClosedTabX: function BMBH_openClosedTabX(index)
	{
		if (index >= 0)
		{
			undoCloseTab(index);
			this._MiscVars.TabImages.splice(index, 1).shift();
		}
	},
	openClosedWindowX: function BMBH_openClosedWindowX(index)
	{
		if (index >= 0)
		{
			undoCloseWindow(index);
		}
	},
	openClosedTabs: function BMBH_openClosedTabs(numTabs)
	{
		for (let i = 0; i < numTabs; i++)
		{
			undoCloseTab(0);
			this._MiscVars.TabImages.splice(0, 1).shift();
		}
	},
	openClosedWindows: function BMBH_openClosedWindows(numWindows)
	{
		for (let i = 0; i < numWindows; i++)
		{
			undoCloseWindow(0);
		}
	},
	openLastClosedTab: function BMBH_openLastClosedTab()
	{
		undoCloseTab(0);
		this._MiscVars.TabImages.splice(0, 1).shift();
	},
	openLastClosedWindow: function BMBH_openLastClosedWindow()
	{
		if (this._MiscVars.canUndoCloseWin)
		{
			undoCloseWindow(0);
		}
	},
	clearClosedTabs: function BMBH_clearClosedTabs()
	{
		var ss = Components.classes["@mozilla.org/browser/sessionstore;1"].getService(Components.interfaces.nsISessionStore);
		if (ss.getClosedTabCount(window) == 0)
		{
			return;
		}
		
		this.PrefsIFace = "browser.sessionstore.";
		try
		{
			const USER_MAX_TABS_UNDO = this.PrefsIFace.getIntPref("max_tabs_undo");
		}
		catch(e){}
		
		this.PrefsIFace.setIntPref("max_tabs_undo", 0);	// causes sessionstore to remove all stored tabs
		if (USER_MAX_TABS_UNDO != null)			// set the value back to the previous value
		{
			this.PrefsIFace.setIntPref("max_tabs_undo", USER_MAX_TABS_UNDO);
		}
		else
		{
			this.PrefsIFace.clearUserPref("max_tabs_undo");
		}
		this.PrefsIFace = this.prefRoot;
	},
	clearClosedWindows: function BMBH_clearClosedWindows()
	{
		var ss = Components.classes["@mozilla.org/browser/sessionstore;1"].getService(Components.interfaces.nsISessionStore);
		if (ss.getClosedWindowCount(window) == 0)
		{
			return;
		}
		
		this.PrefsIFace = "browser.sessionstore.";
		try
		{
			const USER_MAX_WINDOWS_UNDO = this.PrefsIFace.getIntPref("max_windows_undo");
		}
		catch(e){}
		
		this.PrefsIFace.setIntPref("max_windows_undo", 0);	// causes sessionstore to remove all stored windows
		if (USER_MAX_WINDOWS_UNDO != null)			// set the value back to the previous value
		{
			this.PrefsIFace.setIntPref("max_windows_undo", USER_MAX_WINDOWS_UNDO);
		}
		else
		{
			this.PrefsIFace.clearUserPref("max_windows_undo");
		}
		this.PrefsIFace = this.prefRoot;
	},
	/**
		saves the image of the tab by imitating the functionality of nsISessionStore.onTabClose()
		http://mxr.mozilla.org/firefox/source/browser/components/sessionstore/src/nsSessionStore.js#569
	*/
	tabClosed: function BMBH_tabClosed(aEvent)
	{
		bmbextHistory.PrefsIFace = "browser.sessionstore.";
		const USER_MAX_TABS_UNDO = bmbextHistory.PrefsIFace.getIntPref("max_tabs_undo")
		if (USER_MAX_TABS_UNDO == 0)
		{
			return;
		}
		
		var tab = aEvent.originalTarget;
		var linkedPanel = tab.ownerDocument.getElementById(tab.linkedPanel);
		var linkedPanelSSData = linkedPanel.__SS_data;
		var tabbrowser = linkedPanel.getElementsByTagName("browser");
		if (tabbrowser.length > 0)
		{
			tabbrowser = tabbrowser[0];
		}
		else
		{
			tabbrowser = linkedPanel.getElementsByTagName("xul:browser")[0];
		}
		
		if ((tabbrowser.sessionHistory.count > 1 || tabbrowser.currentURI.spec != "about:blank") || (linkedPanelSSData && (linkedPanelSSData.entries.length > 1 || linkedPanelSSData.entries[0].url != "about:blank")))
		{
			bmbextHistory._MiscVars.TabImages.unshift(tab.getAttribute("image")+"");
		}
		var length = bmbextHistory._MiscVars.TabImages.length;
		if (length > USER_MAX_TABS_UNDO)
		{
			bmbextHistory._MiscVars.TabImages.splice(USER_MAX_TABS_UNDO, length - USER_MAX_TABS_UNDO);
		}
	}
}

window.addEventListener("load", bmbextHistory, false);
window.addEventListener("unload", bmbextHistory, false);