/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Bookmarks Menu Button.
 *
 * The Initial Developer of the Original Code is Ronny Perinke.
 * Portions created by the Initial Developer are Copyright (C) 2004-2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 *   Ronny Perinke <www.sephiroth-j.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
 
 /*
#########################
#		new sorting stuff		#
#########################
*/

const bmbextSort = {
 	//#######################################
	//########	private functions and properties	########
	//########						########
	//#######################################
	_Prefs:
	{
		sortrecursive: true,			// include subfolders?
		excludetoolbar: true,			// do not sort the bookmarks toolbar?
		alphabetic: 1,				// 1==ABC-style; -1==ZYX-style
		sortcriterion: 1,				// 1==name; 2==last visit; 3==URL
		sortfolderstyle: 1,			// 1==folders on top; 2==mixed; 3==above
		showextrasortmenu: false		// display full sort menu?
	},
	_MiscVars:
	{
		selection: null,				// current selected object (usually the one on which the right-click was performed)
		PrefsIFace: null,				// refers to the nsIPref interface
	},
	get prefRoot()
	{
		return "BMBExt.Sort.";
	},
	get StringBundle()
	{
		var stringBundle = document.getElementById("bmbextstr");
		delete this.StringBundle;
		this.StringBundle = stringBundle;
		return stringBundle;
	},
	get PrefsIFace()
	{
		if (!this._MiscVars.PrefsIFace)
		{
			this.PrefsIFace = this.prefRoot;
			this.__defineGetter__("PrefsIFace", function() { return this._Misc.PrefsIFace; });
		}
		return this._MiscVars.PrefsIFace;
	},
	set PrefsIFace(aBranch)
	{
		if (typeof(aBranch) != "string")
		{
			throw Components.results.NS_ERROR_ILLEGAL_VALUE + "\nValue must be a string type";
		}
		
		this._MiscVars.PrefsIFace = Components.classes["@mozilla.org/preferences;1"].getService(Components.interfaces.nsIPref);
		this._MiscVars.PrefsIFace = this._MiscVars.PrefsIFace.getBranch((aBranch != "")?aBranch:this.prefRoot);
	},
	get selection()
	{
		return this._MiscVars.selection;
	},
	set selection(aObject)
	{
		this._MiscVars.selection = aObject;
	},
	//#######################################
	//########	/private functions and properties	########
	//########						########
	//#######################################
	
	handleEvent: function BMBS_handleEvent(aEvent)
	{
		switch(aEvent.type)
		{
			case "load":
				this.init();
				break;
			
			case "unload":
				this.quit();
				break;
				
			case "popupshowing":
				this.onContextShowing_Places(aEvent);
				break;
			
			case "popuphiding":
				this.onContextHiding_Places(aEvent);
				break;
			
			case "popuphidden":
				this.onContextHidden_Places(aEvent);
				break;
		}
	},
	
	init: function BMBS_init()
	{
		this.loadPrefs();
		this.addSortSupport_Places();
	},
	quit: function BMBS_quit()
	{
		this._MiscVars.PrefsIFace = null;
		this._MiscVars.selection = null;
		delete this.StringBundle;
	},
	
	loadPrefs: function BMBS_loadPrefs()
	{
		this.PrefsIFace = this.prefRoot;
		
		this._Prefs.sortrecursive = this.PrefsIFace.getBoolPref("settings.sortrecursive");
		this._Prefs.excludetoolbar = this.PrefsIFace.getBoolPref("settings.excludetoolbar");
		this._Prefs.sortfolderstyle = this.PrefsIFace.getIntPref("settings.sortfolderstyle");
		this._Prefs.alphabetic = this.PrefsIFace.getIntPref("settings.alphabetic");
		this._Prefs.sortcriterion = this.PrefsIFace.getIntPref("settings.sortcriterion");
		this._Prefs.showextrasortmenu = this.PrefsIFace.getBoolPref("settings.showextrasortmenu");
	},
	
	/**
		PLACES STUFF
	*/
	addSortSupport_Places: function BMBS_addSortSupport_Places()
	{
		var items = document.getElementsByAttribute("place", "*");
		for (let i=0, l=items.length; i<l; i++)
		{
			items[i].controllers.appendController(new BMBSCmdController(items[i]));
		}
		document.getElementById("placesCommands").addEventListener("commandupdate", BMBSC_goUpdateCommand, false);
		
		var contextmenu = document.getElementById("placesContext");
		
		// add support for additional sort menu
		contextmenu.addEventListener("popupshowing", this, true);
		
		// prevent calling PlacesMenuDNDController._closePopups() in normal bookmarks menu
		contextmenu.addEventListener("popuphidden", this, true);
	},
	
	onContextShowing_Places: function BMBS_onContextShowing_Places(aEvent)
	{
		var view = PlacesUIUtils.getViewForNode(document.popupNode);
		this.selection = view.selectedNode;
		
		if (aEvent.target.id == "placesContext")
		{
			this.addExtraContextMenu_Places(aEvent.target);
		}
	},
	onContextHiding_Places: function BMBS_onContextHiding_Places(aEvent)
	{
		if (aEvent.target.id != "placesContext")
		{
			// do not close popups, when sortmenu popup closes
			aEvent.stopPropagation();
		}
	},
	onContextHidden_Places: function BMBS_onContextHidden_Places(aEvent)
	{
		if (aEvent.target.id != "placesContext")
		{
			// do not close popups, when sortmenu popup closes
			// actually, it's not required anymore after fix for bug #195031 has landed
			aEvent.stopPropagation();
		}
	},
	
	addExtraContextMenu_Places: function BMBS_addExtraContextMenu_Places(placesContext)
	{
		var sortmenu = document.getElementById("bmbext-SortMenu");
		if (!this.selection || PlacesUtils.nodeIsLivemarkItem(this.selection) || PlacesUtils.nodeIsLivemarkContainer(this.selection) || PlacesUtils.nodeIsReadOnly(this.selection) || (!PlacesUtils.nodeIsBookmark(this.selection) && !PlacesUtils.nodeIsFolder(this.selection)))
		{
			if (!sortmenu.hidden)
			{
				sortmenu.hidden = true;
			}
			return;
		}
		
		if (!sortmenu.hidden && !this._Prefs.showextrasortmenu)
		{
			sortmenu.hidden = true;
			return;
		}
		if (sortmenu.hidden && this._Prefs.showextrasortmenu)
		{
			sortmenu.hidden = false;
		}
	},
	
	PlacesSortFolderByXXX: function BMBS_PlacesSortFolderByXXX(aSortType, aSortDir)
	{
		if (!aSortDir)
		{
			aSortDir = this._Prefs.alphabetic * this._Prefs.sortfolderstyle;
		}
		if (!aSortType)
		{
			switch(this._Prefs.sortcriterion)
			{
				case 1:
				default:
					aSortType = "Name";
					break;
				case 2:
					aSortType = "LastVisit";
					break;
				case 3:
					aSortType = "URL";
					break;
			}
		}
		
		var selectedNode = this.selection;
		if (PlacesUtils.nodeIsBookmark(selectedNode))
		{
			selectedNode = selectedNode.parent;
		}
		var txn = new BMBSortFolderByXXXTransaction(selectedNode.itemId, selectedNode.bookmarkIndex, aSortType, aSortDir, this._Prefs.sortrecursive, this._Prefs.excludetoolbar);
		PlacesUIUtils.ptm.doTransaction(txn);
	}
}

/**
* Sort a folder by XXX
*/
function BMBSortFolderByXXXTransaction(aFolderId, aFolderIndex, aSortType, aSortDir, SortRecursive, SortExcludePersonalToolbar)
{
	this._folderId = aFolderId;
	this._folderIndex = aFolderIndex;
	this._oldOrder = null,
	this.redoTransaction = this.doTransaction;
	this.sortType = aSortType;
	this.sortDir = aSortDir;
	this.sortDirAbs = Math.abs(aSortDir);
	this.recursive = SortRecursive;
	this.excludePersonalToolbar = SortExcludePersonalToolbar;
}
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
BMBSortFolderByXXXTransaction.prototype = {
	/*
		Base Transaction Stuff
	*/
	// for child-transactions
	get wrappedJSObject()
	{
		return this;
	},
	// nsITransaction
	redoTransaction: function BMBSFBX_redoTransaction()
	{
		throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
	},
	get isTransient()
	{
		return false;
	},
	merge: function BMBSFBX_merge(transaction)
	{
		return false;
	},
	// nsISupports
	QueryInterface: XPCOMUtils.generateQI([Components.interfaces.nsITransaction, Components.interfaces.nsISupports]),
	
	/*
		Sort Implementation
	*/
	doTransaction: function BMBSFBX_doTransaction()
	{
		if (PlacesUtils.nodeIsLivemarkContainer(this._folderId) || (this.excludePersonalToolbar && PlacesUtils.bookmarks.toolbarFolder == this._folderId))
		{
			return;
		}
		this._oldOrder = [];
		var newOrder = [];
		//var contents = PlacesUtils.getFolderContents(this._folderId, false, false);
		var contents = PlacesUtils.getFolderContents(this._folderId, false, false).root;
		var count = contents.childCount;
		for (var item, rectxn, i = 0; i < count; i++)
		{
			item = contents.getChild(i);
			if (!PlacesUtils.nodeIsLivemarkItem(item))	// exclude livemarks
			{
				if (this.recursive && PlacesUtils.nodeIsFolder(item) && !PlacesUtils.nodeIsLivemarkContainer(item))
				{
					rectxn = new BMBSortFolderByXXXTransaction(item.itemId, item.bookmarkIndex, this.sortType, this.sortDir, this.recursive, this.excludePersonalToolbar);
					PlacesUIUtils.ptm.doTransaction(rectxn);
				}
				this._oldOrder[item.itemId] = i;
				newOrder.push(item);
			}
		}
		dump("anzahl: " + count + "\n");
		
		const self = this;
		function sortItemsFunc(a, b)
		{
			var aisfolder = PlacesUtils.nodeIsContainer(a);
			var bisfolder = PlacesUtils.nodeIsContainer(b);
			var atitle, btitle;
			
			switch(self.sortType)
			{
				case "Name":
				default:
					atitle = a.title;
					btitle = b.title;
					break;
				case "LastVisit":
					atitle = a.time;
					btitle = b.time;
					break;
				case "URL":
					atitle = a.uri;
					btitle = b.uri;
					break;
			}
			
			// sort folders
			switch(self.sortDirAbs)
			{
				case 1:					//place folders above bookmarks
				default:
					if (aisfolder && !bisfolder)
						return -1;
					if (bisfolder && !aisfolder)
						return 1;
					break;
					
				case 2:					//mix folders and bookmarks for sortbyname and sortbylastvisit
					break;
					
				case 3:					//place folders below bookmarks
					if (aisfolder && !bisfolder)
						return 1;
					if (bisfolder && !aisfolder)
						return -1;
					break;
			}
			
			// sort other items
			if (atitle == btitle)
			{
				return 0;
			}
			else
			{
				if (self.sortDir >= 0)
				{
					if (self.sortType != "LastVisit")
					{
						// ABC
						//return (atitle < btitle) ? -1 : 1;
						return atitle.localeCompare(btitle);
					}
					else
					{
						// newest first
						return (atitle > btitle) ? -1 : 1;
					}
				}
				else
				{
					if (self.sortType != "LastVisit")
					{
						// ZYX
						//return (atitle < btitle) ? 1 : -1;
						return btitle.localeCompare(atitle);
					}
					else
					{
						// newest last
						return (atitle > btitle) ? 1 : -1;
					}
				}
			}
		}
		newOrder.sort(sortItemsFunc);
		
		var callback = {
			runBatched: function()
			{
				for (let i = 0; i < count; i++)
				{
					PlacesUtils.bookmarks.setItemIndex(newOrder[i].itemId, i);
				}
			}
		}
		PlacesUtils.bookmarks.runInBatchMode(callback, null);
	},
	undoTransaction: function BMBSFBX_undoTransaction()
	{
		var callback = {
			self: this,
			runBatched: function()
			{
				for (item in self._oldOrder)
				{
					PlacesUtils.bookmarks.setItemIndex(item, self._oldOrder[item]);
				}
			}
		}
		PlacesUtils.bookmarks.runInBatchMode(callback, null);
	}
}

function BMBSCmdController(aView)
{
	this._view = aView;
}
BMBSCmdController.prototype = {
	_view: null,
	onEvent: function BMBSC_onEvent(eventName)
	{
	},
	doCommand: function BMBSC_doCommand(aCmd)
	{
		bmbext.dump(":BMBSC::doCommand: "+aCmd);
		switch(aCmd)
		{
			case "bmbCmd:sortnow":
				bmbextSort.PlacesSortFolderByXXX();
				this._view.place = this._view.place;
				break;
		}
	},
	supportsCommand: function BMBSC_supportsCommand(aCmd)
	{
		return (aCmd == "bmbCmd:sortnow");
	},
	isCommandEnabled: function BMBSC_isCommandEnabled(aCmd)
	{
		bmbext.dump(":BMBSC::isCommandEnabled: "+aCmd);
		switch(aCmd)
		{
			case "bmbCmd:sortnow":
				var selectedNode = this._view.selectedNode;
				return selectedNode && !PlacesUtils.nodeIsReadOnly(selectedNode) && (PlacesUtils.nodeIsFolder(selectedNode) || PlacesUtils.nodeIsBookmark(selectedNode));
				
			default:
				return false;
		}
	}
}

function BMBSC_goUpdateCommand(aEvent)
{
	goUpdateCommand("bmbCmd:sortnow");
}

window.addEventListener("load", bmbextSort, false);
window.addEventListener("unload", bmbextSort, false);